@@ -149,6 +149,7 @@ front_arg_new(void)
 	fa->sni_nomatch_abort = -1;
 	fa->selected_protos = 0;
 	fa->prefer_server_ciphers = -1;
+	fa->client_verify = -1;
 
 	return (fa);
 }
@@ -162,7 +163,8 @@ front_arg_destroy(struct front_arg *fa)
 	free(fa->ip);
 	free(fa->port);
 	free(fa->pspec);
-	free(fa->ciphers);
+	free(fa->ciphers_tlsv12);
+	free(fa->ciphersuites_tlsv13);
 	HASH_ITER(hh, fa->certs, cf, cftmp) {
 		CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
 		HASH_DEL(fa->certs, cf);
@@ -201,7 +203,7 @@ config_new(void)
 	r->BACK_IP			= strdup("127.0.0.1");
 	r->BACK_PORT			= strdup("8000");
 	r->NCORES			= 1;
-	r->CIPHER_SUITE			= strdup(CFG_DEFAULT_CIPHERS);
+	r->CIPHERS_TLSv12		= strdup(CFG_DEFAULT_CIPHERS);
 	r->ENGINE			= NULL;
 	r->BACKLOG			= 100;
 	r->SNI_NOMATCH_ABORT		= 0;
@@ -215,6 +217,8 @@ config_new(void)
 	r->OCSP_RESP_TMO		= 10.0;
 	r->OCSP_CONN_TMO		= 4.0;
 	r->OCSP_REFRESH_INTERVAL	= 1800;
+	r->CLIENT_VERIFY		= SSL_VERIFY_NONE;
+	r->CLIENT_VERIFY_CA		= NULL;
 #ifdef TCP_FASTOPEN_WORKS
 	r->TFO				= 0;
 #endif
@@ -231,7 +235,7 @@ config_new(void)
 	r->SHCUPD_MCASTTTL		= NULL;
 #endif
 
-	r->LOG_LEVEL			= 0;
+	r->LOG_LEVEL			= 1;
 	r->SYSLOG			= 0;
 	r->SYSLOG_FACILITY		= LOG_DAEMON;
 	r->TCP_KEEPALIVE_TIME		= 3600;
@@ -290,7 +294,8 @@ config_destroy(hitch_config *cfg)
 	if (cfg->CERT_DEFAULT != NULL)
 		cfg_cert_file_free(&cfg->CERT_DEFAULT);
 
-	free(cfg->CIPHER_SUITE);
+	free(cfg->CIPHERS_TLSv12);
+	free(cfg->CIPHERSUITES_TLSv13);
 	free(cfg->ENGINE);
 	free(cfg->PIDFILE);
 	free(cfg->OCSP_DIR);
@@ -298,6 +303,7 @@ config_destroy(hitch_config *cfg)
 	free(cfg->ALPN_PROTOS_LV);
 	free(cfg->PEM_DIR);
 	free(cfg->PEM_DIR_GLOB);
+	free(cfg->CLIENT_VERIFY_CA);
 #ifdef USE_SHARED_CACHE
 	int i;
 	free(cfg->SHCUPD_IP);
@@ -841,7 +847,7 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		cfg->SELECTED_TLS_PROTOS = SSL_OPTION_PROTOS;
 	} else if (strcmp(k, CFG_CIPHERS) == 0) {
 		if (strlen(v) > 0) {
-			config_assign_str(&cfg->CIPHER_SUITE, v);
+			config_assign_str(&cfg->CIPHERS_TLSv12, v);
 		}
 	} else if (strcmp(k, CFG_SSL_ENGINE) == 0) {
 		if (strlen(v) > 0) {
@@ -1230,7 +1236,7 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "\n");
 	fprintf(out, "      --tls                   TLSv1 (default. No SSLv3)\n");
 	fprintf(out, "      --ssl                   SSLv3 (enables SSLv3)\n");
-	fprintf(out, "  -c  --ciphers=SUITE         Sets allowed ciphers (Default: \"%s\")\n", config_disp_str(cfg->CIPHER_SUITE));
+	fprintf(out, "  -c  --ciphers=SUITE         Sets allowed ciphers (Default: \"%s\")\n", config_disp_str(cfg->CIPHERS_TLSv12));
 	fprintf(out, "  -e  --ssl-engine=NAME       Sets OpenSSL engine (Default: \"%s\")\n", config_disp_str(cfg->ENGINE));
 	fprintf(out, "  -O  --prefer-server-ciphers Prefer server list order\n");
 	fprintf(out, "\n");
@@ -1383,6 +1389,7 @@ int
 config_parse_cli(int argc, char **argv, hitch_config *cfg)
 {
 	static int tls = 0, ssl = 0;
+	struct front_arg *fa, *fatmp;
 	static int client = 0;
 	int c, i;
 
@@ -1572,6 +1579,26 @@ CFG_ON('s', CFG_SYSLOG);
 		return (1);
 	}
 
+	if (cfg->CLIENT_VERIFY != SSL_VERIFY_NONE &&
+	    cfg->CLIENT_VERIFY_CA == NULL) {
+		config_error_set("Setting 'client-verify-ca' is required when"
+		    " configuring client-verify");
+		return (1);
+	}
+
+	HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp) {
+		if (fa->client_verify != -1 &&
+		    fa->client_verify != SSL_VERIFY_NONE) {
+			if (!fa->client_verify_ca && !cfg->CLIENT_VERIFY_CA) {
+				config_error_set("No 'client-verify-ca' "
+				    "configured for frontend '%s'",
+				    fa->pspec);
+				return (1);
+			}
+		}
+
+	}
+
 
 #ifdef USE_SHARED_CACHE
 	if (cfg->SHCUPD_IP != NULL && ! cfg->SHARED_CACHE) {
@@ -1654,7 +1681,6 @@ CFG_ON('s', CFG_SYSLOG);
 	}
 
 	if (cfg->PMODE == SSL_SERVER && cfg->CERT_DEFAULT == NULL) {
-		struct front_arg *fa, *fatmp;
 		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp)
 			if (HASH_CNT(hh, fa->certs) == 0) {
 				config_error_set("No x509 certificate PEM file "
