@@ -37,6 +37,8 @@
 
 #include "config.h"
 
+#include <openssl/x509.h>
+#include <openssl/x509_vfy.h>
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -330,17 +332,28 @@ init_dh(SSL_CTX *ctx, const char *cert)
 	return (0);
 }
 
-static int init_ecdh(SSL_CTX *ctx) {
+static int init_ecdh(SSL_CTX *ctx, const char *ecdh_curve) {
 #ifndef OPENSSL_NO_EC
-#ifdef NID_X9_62_prime256v1
-	EC_KEY *ecdh = NULL;
-	ecdh = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
-	SSL_CTX_set_tmp_ecdh(ctx, ecdh);
-	EC_KEY_free(ecdh);
-	LOG("{core} ECDH Initialized with NIST P-256\n");
-#endif /* NID_X9_62_prime256v1 */
+	if (ecdh_curve == NULL || strcmp(ecdh_curve, "auto") == 0) {
+		/* For openssl >= 1.1 'auto' is default on and this is
+		 * just a NOOP macro. */
+		AN(SSL_CTX_set_ecdh_auto(ctx, 1));
+	} else {
+		/* For openssl >= 1.1 this should really be "groups"
+		 * instead of "curves", but we use the old name to
+		 * support openssl 1.0 as well */
+		if (SSL_CTX_set1_curves_list(ctx,
+			ecdh_curve) == 0) {
+			log_ssl_error(NULL,
+			    "{core} Configuring ecdh curves '%s' failed\n",
+			    ecdh_curve);
+		}
+	}
+	LOG("{core} ECDH Initialized\n");
+#else
+	(void) ctx;
+	(void) ecdh_curve;
 #endif /* OPENSSL_NO_EC */
-
 	return (0);
 }
 #endif /* OPENSSL_NO_DH */
@@ -892,7 +905,74 @@ Find_issuer(X509 *subj, STACK_OF(X509) *chain)
        return (NULL);
 }
 
+static int
+client_vfy_cb(int preverify_ok, X509_STORE_CTX *storectx)
+{
+	proxystate *ps;
+	SSL *ssl;
 
+	ssl = X509_STORE_CTX_get_ex_data(storectx,
+	    SSL_get_ex_data_X509_STORE_CTX_idx());
+	CAST_OBJ_NOTNULL(ps, SSL_get_app_data(ssl), PROXYSTATE_MAGIC);
+	if (preverify_ok)
+		ps->client_cert_conn = 1;
+
+	return (preverify_ok);
+}
+
+static int
+client_vfy_init(SSL_CTX *ctx, int flags, const char *cafile)
+{
+	X509_STORE *vfy;
+	STACK_OF(X509_OBJECT) *objs;
+	X509_OBJECT *o;
+	X509 *crt;
+	int i;
+
+	AN(cafile);
+	assert(flags != SSL_VERIFY_NONE);
+	AN(flags & SSL_VERIFY_PEER);
+
+	vfy = X509_STORE_new();
+	if (!vfy) {
+		log_ssl_error(NULL, "X509_STORE_new: allocation failed");
+		return (1);
+	}
+	if (X509_STORE_load_locations(vfy, cafile, NULL) == 0) {
+		log_ssl_error(NULL, "client_verify_ca: unable to "
+		    "load file '%s'",
+		    cafile);
+		X509_STORE_free(vfy);
+		return (1);
+	}
+
+	SSL_CTX_set1_verify_cert_store(ctx, vfy);
+
+#ifdef HAVE_X509_STORE_GET0_OBJECTS
+	objs = X509_STORE_get0_objects(vfy);
+#else
+	objs = vfy->objs;
+#endif
+	for (i = 0; i < sk_X509_OBJECT_num(objs); i++) {
+		o = sk_X509_OBJECT_value(objs, i);
+#ifdef HAVE_X509_OBJECT_GET0_X509
+		crt = X509_OBJECT_get0_X509(o);
+#else
+		crt = o->data.x509;
+#endif
+		if (crt != NULL) {
+			/* SSL_CTX_add_client_CA makes a copy of the
+			 * subject name, so the X509_STORE_free below
+			 * is safe. */
+			SSL_CTX_add_client_CA(ctx, crt);
+		}
+	}
+
+	SSL_CTX_set_verify(ctx, flags, client_vfy_cb);
+
+	X509_STORE_free(vfy);
+	return (0);
+}
 
 /* Initialize an SSL context */
 static sslctx *
@@ -903,17 +983,23 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 	sslctx *sc;
 	EVP_PKEY *pkey;
 	int selected_protos = CONFIG->SELECTED_TLS_PROTOS;
-	char *ciphers = CONFIG->CIPHER_SUITE;
+	char *ciphers = CONFIG->CIPHERS_TLSv12;
+	char *ciphersuites = CONFIG->CIPHERSUITES_TLSv13;
 	int pref_srv_ciphers = CONFIG->PREFER_SERVER_CIPHERS;
+	int client_verify = CONFIG->CLIENT_VERIFY;
 
 	if (fa != NULL) {
 		CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
 		if (fa->selected_protos != 0)
 			selected_protos = fa->selected_protos;
-		if (fa->ciphers != NULL)
-			ciphers = fa->ciphers;
+		if (fa->ciphers_tlsv12 != NULL)
+			ciphers = fa->ciphers_tlsv12;
 		if (fa->prefer_server_ciphers != -1)
 			pref_srv_ciphers = fa->prefer_server_ciphers;
+		if (fa->ciphersuites_tlsv13)
+			ciphersuites = fa->ciphersuites_tlsv13;
+		if (fa->client_verify != -1)
+			client_verify = fa->client_verify;
 	}
 
 	long ssloptions = SSL_OP_NO_SSLv2 | SSL_OP_ALL |
@@ -961,9 +1047,31 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 		}
 	}
 
+#if HAVE_TLS_1_3
+	if (ciphersuites != NULL) {
+		if (SSL_CTX_set_ciphersuites(ctx, ciphersuites) != 1) {
+			log_ssl_error(NULL, "{core} SSL_CTX_set_ciphersuites");
+			return (NULL);
+		}
+	}
+#else
+	(void) ciphersuites;
+#endif
+	if (client_verify != SSL_VERIFY_NONE) {
+		const char *ca = CONFIG->CLIENT_VERIFY_CA;
+		if (fa && fa->client_verify_ca)
+			ca = fa->client_verify_ca;
+		AN(ca);
+		if (client_vfy_init(ctx, client_verify, ca))
+			return (NULL);
+	}
+
 	if (pref_srv_ciphers)
 		SSL_CTX_set_options(ctx, SSL_OP_CIPHER_SERVER_PREFERENCE);
 
+	AN(SSL_CTX_set_session_id_context(ctx, (const unsigned char *) "hitch",
+		strlen("hitch")));
+
 	ALLOC_OBJ(sc, SSLCTX_MAGIC);
 	AN(sc);
 	sc->filename = strdup(cf->filename);
@@ -1016,7 +1124,7 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 
 #ifndef OPENSSL_NO_DH
 	init_dh(ctx, cf->filename);
-	init_ecdh(ctx);
+	init_ecdh(ctx, CONFIG->ECDH_CURVE);
 #endif /* OPENSSL_NO_DH */
 
 #ifndef OPENSSL_NO_TLSEXT
@@ -1037,10 +1145,15 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 	if (CONFIG->OCSP_DIR) {
 		char *fn = HOCSP_fn(sc->filename);
 		/* attempt loading of cached ocsp staple */
-		if (fn != NULL && HOCSP_init_file(fn, sc, 1) == 0) {
-			LOG("{core} Loaded cached OCSP staple for cert '%s'\n",
-			    sc->filename);
-			sc->staple_fn = fn;
+		if (fn != NULL) {
+			if (HOCSP_init_file(fn, sc, 1) == 0) {
+				LOG("{core} Loaded cached OCSP staple "
+				    "for cert '%s'\n", sc->filename);
+				sc->staple_fn = fn;
+			}
+			else {
+				free(fn);
+			}
 		}
 	}
 
@@ -1984,16 +2097,30 @@ write_proxy_v2(proxystate *ps, const struct sockaddr *local)
 		}
 	}
 	if (CONFIG->PROXY_TLV) {
-		char *tlvp = base + len;
+		X509 *crt;
 		ssize_t sz = 0;
+		struct pp2_tlv_ssl *tlv;
+		char *tlvp = base + len;
 		const char *tmp;
 
 		tlvp[0] = PP2_TYPE_SSL;
 		/* tlvp[1..2] to be updated with payload length later */
 		len += 3;
+		tlv = (struct pp2_tlv_ssl *) (&base[len]);
+		tlv->client = PP2_CLIENT_SSL;
+		tlv->verify = 1;
 
-		base[len] = PP2_CLIENT_SSL;
-		base[len + 1] = 1;
+		/* PP2_CLIENT_CERT_SESS */
+		crt = SSL_get_peer_certificate(ps->ssl);
+		if (crt) {
+			tlv->client |= PP2_CLIENT_CERT_SESS;
+			tlv->verify = htonl(SSL_get_verify_result(ps->ssl));
+			X509_free(crt);
+		}
+
+		/* PP2_CLIENT_CERT_CONN */
+		if (ps->client_cert_conn)
+			tlv->client |= PP2_CLIENT_CERT_CONN;
 		len += 5;
 		sz += 5;
 
@@ -3335,62 +3462,6 @@ init_signals()
 
 }
 
-#define NULL_DEV "/dev/null"
-static void
-daemonize()
-{
-	/* logging.c */
-	if (logfile == stdout || logfile == stderr) {
-		logfile = NULL;
-	}
-
-	/* go to root directory */
-	if (chdir("/") != 0) {
-		ERR("Unable change directory to /: %s\n", strerror(errno));
-		exit(1);
-	}
-
-	/* let's make some children, baby :) */
-	pid_t pid = fork();
-	if (pid < 0) {
-		ERR("Unable to daemonize: fork failed: %s\n", strerror(errno));
-		exit(1);
-	}
-
-	/* am i the parent? */
-	if (pid != 0) {
-		LOGL("{core} Daemonized as pid %d.\n", pid);
-		exit(0);
-	}
-
-	/* reopen standard streams to null device */
-	if (freopen(NULL_DEV, "r", stdin) == NULL) {
-		ERR("Unable to reopen stdin to %s: %s\n",
-		    NULL_DEV, strerror(errno));
-		exit(1);
-	}
-	if (freopen(NULL_DEV, "w", stdout) == NULL) {
-		ERR("Unable to reopen stdout to %s: %s\n",
-		    NULL_DEV, strerror(errno));
-		exit(1);
-	}
-	if (freopen(NULL_DEV, "w", stderr) == NULL) {
-		ERR("Unable to reopen stderr to %s: %s\n",
-		    NULL_DEV, strerror(errno));
-		exit(1);
-	}
-
-	/* this is child, the new master */
-	pid_t s = setsid();
-	if (s < 0) {
-		ERR("Unable to create new session, setsid(2) failed: "
-		    "%s :: %d\n", strerror(errno), s);
-		exit(1);
-	}
-
-	LOG("Successfully daemonized as pid %d.\n", getpid());
-}
-
 static void
 openssl_check_version()
 {
@@ -3924,17 +3995,6 @@ main(int argc, char **argv)
 		return (1);
 	}
 
-	if (CONFIG->TEST) {
-		fprintf(stderr, "Trying to initialize SSL contexts with your"
-		    " certificates\n");
-		init_globals();
-		init_openssl();
-		init_certs();
-		fprintf(stderr, "%s configuration looks ok.\n",
-		    basename(argv[0]));
-		return (0);
-	}
-
 	if (CONFIG->LOG_FILENAME) {
 		FILE* f;
 		if ((f = fopen(CONFIG->LOG_FILENAME, "a")) == NULL) {
@@ -3955,8 +4015,18 @@ main(int argc, char **argv)
 	}
 	AZ(setvbuf(logfile, NULL, _IONBF, BUFSIZ));
 
-	if (CONFIG->DAEMONIZE && (logfile == stdout || logfile == stderr))
-		logfile = NULL;
+	if (CONFIG->TEST) {
+		/* Override log level for config test */
+		CONFIG->LOG_LEVEL = 3;
+		fprintf(stderr, "Trying to initialize SSL contexts with your"
+		    " certificates\n");
+		init_globals();
+		init_openssl();
+		init_certs();
+		fprintf(stderr, "%s configuration looks ok.\n",
+		    basename(argv[0]));
+		return (0);
+	}
 
 	LOGL("{core} %s starting\n", PACKAGE_STRING);
 	create_workers = 1;
@@ -3998,8 +4068,19 @@ main(int argc, char **argv)
 		exit(1);
 	}
 
-	if (CONFIG->DAEMONIZE)
-		daemonize();
+	if (CONFIG->DAEMONIZE) {
+		if (!CONFIG->SYSLOG && !CONFIG->LOG_FILENAME) {
+			LOG("{core} Warning: daemonizing with neither "
+			    "'syslog' nor 'log-filename' configured: "
+			    "Hitch will not produce log messages.\n");
+		}
+		if (logfile == stdout || logfile == stderr)
+			logfile = NULL;
+		if (daemon(0, 0) == -1) {
+			ERR("Unable to daemonize: %s\n", strerror(errno));
+			exit(1);
+		}
+	}
 
 	master_pid = getpid();
 
